                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/

#ifndef _RESOLUTION_H_
#define _RESOLUTION_H_

#include "Common/Set.h"

bool setResolution(unsigned int width, unsigned int height, unsigned int bpp = 32, unsigned int hz = 85){
	DEVMODE *mode = new DEVMODE;
	Set <DEVMODE *> modes;
	
	// Collect all displaymodes
	int i = 0;
	while (EnumDisplaySettings(NULL, i, mode) != 0){
		modes.add(mode);
		mode = new DEVMODE;
		i++;
	}
	delete mode;
	
	// Find the best color depth
	bool existBpp = false;
	unsigned int closestAbove = 0xFFFFFFFF, closestBelow = 0;
	for (i = 0; i < modes.getSize(); i++){
		mode = modes[i];
		if (existBpp){
			if (mode->dmBitsPerPel != bpp){
				delete mode;
				modes.remove(i);
				i--;
			}
		} else {
			if (mode->dmBitsPerPel == bpp){
				existBpp = true;
				i = -1;
			} else if (mode->dmBitsPerPel > bpp){
				if (mode->dmBitsPerPel < closestAbove) closestAbove = mode->dmBitsPerPel;
			} else {
				if (mode->dmBitsPerPel > closestBelow) closestBelow = mode->dmBitsPerPel;
			}
		}
	}
	
	// Filter out unwanted color depths
	if (!existBpp){
		for (i = 0; i < modes.getSize(); i++){
			mode = (DEVMODE *) modes[i];
			if (closestAbove != 0xFFFFFFFF){
				if (mode->dmBitsPerPel != closestAbove){
					delete mode;
					modes.remove(i);
					i--;
				}
			} else {
				if (mode->dmBitsPerPel != closestBelow){
					delete mode;
					modes.remove(i);
					i--;
				}
			}
		}
	}
	
	// Find closest width
	unsigned int closestWidth = 0;
	for (i = 0; i < modes.getSize(); i++){
		mode = (DEVMODE *) modes[i];
		if (abs(mode->dmPelsWidth - width) < abs(closestWidth - width)) closestWidth = mode->dmPelsWidth;
	}
	
	// Filter out unwanted widths
	for (i = 0; i < modes.getSize(); i++){
		mode = (DEVMODE *) modes[i];
		if (abs(mode->dmPelsWidth != closestWidth)){
			delete mode;
			modes.remove(i);
			i--;
		}
	}
	
	// Find closest height
	unsigned int closestHeight = 0;
	for (i = 0; i < modes.getSize(); i++){
		mode = (DEVMODE *) modes[i];
		if (abs(mode->dmPelsHeight - height) < abs(closestHeight - height)) closestHeight = mode->dmPelsHeight;
	}
	
	// Filter out unwanted heights
	for (i = 0; i < modes.getSize(); i++){
		mode = (DEVMODE *) modes[i];
		if (abs(mode->dmPelsHeight != closestHeight)){
			delete mode;
			modes.remove(i);
			i--;
		}
	}
	
	// Find closest refreshrate
	unsigned int closestHz = 0;
	for (i = 0; i < modes.getSize(); i++){
		mode = (DEVMODE *) modes[i];
		if (abs(mode->dmDisplayFrequency - hz) < abs(closestHz - hz)) closestHz = mode->dmDisplayFrequency;
	}
	
	// Filter out unwanted refreshrates
	for (i = 0; i < modes.getSize(); i++){
		mode = (DEVMODE *) modes[i];
		if (abs(mode->dmDisplayFrequency != closestHz)){
			delete mode;
			modes.remove(i);
			i--;
		}
	}
	
	// Hopefully we got something left :)
	if (modes.getSize() > 0){
		mode = (DEVMODE *) modes[0];
		if (ChangeDisplaySettings(mode, CDS_FULLSCREEN) == DISP_CHANGE_SUCCESSFUL){
			for (i = 0; i < modes.getSize(); i++)
				delete modes[i];
			return true;
		}
	}
	
	for (i = 0; i < modes.getSize(); i++)
		delete modes[i];
	
	return false;
}

bool resetResolution(){
	ChangeDisplaySettings(NULL, 0);
	return true;
}

// Using Direct3D
#ifdef _D3D8_H_

bool d3dFindBestResolution(int &width, int &height, D3DFORMAT *formats, int nFormats, int &refresh){
	Set <D3DDISPLAYMODE> modes;
	D3DDISPLAYMODE mode;
	int i, j, len = d3d->GetAdapterModeCount(D3DADAPTER_DEFAULT);

	// Collect every mode with an acceptable format
	j = 0;
	do {
		for (i = 0; i < len; i++){
			d3d->EnumAdapterModes(D3DADAPTER_DEFAULT, i, &mode);
			if (mode.Format == formats[j]){
				modes.add(mode);
			}
		}
		j++;
	} while (j < nFormats && modes.getSize() == 0);


	// Find closest width
	unsigned int closestWidth = 0;
	len = modes.getSize();
	for (i = 0; i < len; i++){
		if (abs(modes[i].Width - width) < abs(closestWidth - width)) closestWidth = modes[i].Width;
	}
	
	// Filter out unwanted widths
	for (i = 0; i < modes.getSize(); i++){
		if (abs(modes[i].Width != closestWidth)){
			modes.remove(i);
			i--;
		}
	}
	
	// Find closest height
	unsigned int closestHeight = 0;
	len = modes.getSize();
	for (i = 0; i < len; i++){
		if (abs(modes[i].Height - height) < abs(closestHeight - height)) closestHeight = modes[i].Height;
	}
	
	// Filter out unwanted heights
	for (i = 0; i < modes.getSize(); i++){
		if (abs(modes[i].Height != closestHeight)){
			modes.remove(i);
			i--;
		}
	}

	unsigned int closestRefresh = 0;
	len = modes.getSize();
	for (i = 0; i < len; i++){
		if (abs(modes[i].RefreshRate - refresh) < abs(closestRefresh - refresh)) closestRefresh = modes[i].RefreshRate;
	}
	
	// Filter out unwanted Refreshs
	for (i = 0; i < modes.getSize(); i++){
		if (abs(modes[i].RefreshRate != closestRefresh)){
			modes.remove(i);
			i--;
		}
	}

	if (modes.getSize() > 0){
		width    = modes[0].Width;
		height   = modes[0].Height;
		*formats = modes[0].Format;
		refresh  = modes[0].RefreshRate;
		return true;
	} else return false;
}

#endif // _D3D8_H_


#endif // _RESOLUTION_H_