#ifndef _NETWORK_H_
#define _NETWORK_H_

#include <winsock.h>

bool initSockets(){
	WSADATA wsaData;
	
	if (WSAStartup(MAKEWORD(1, 1), &wsaData) != 0){
		//printf("initSockets() error\n");
		//exit(1);
		return false;
	}
	
	if (LOBYTE(wsaData.wVersion) != 1 ||
        HIBYTE(wsaData.wVersion) != 1) {
		//printf("initSockets() error\n");
		WSACleanup();
		//exit(1);
		return false;
	}

	return true;
}

void closeSockets(){
	WSACleanup();
}

/* ################################################################################################### */

class Server {
protected:
	static SOCKET serverSocket;
	static sockaddr_in server;

	SOCKET clientSocket;
	sockaddr_in client;

	bool isConnected;

public:
	Server(){
		isConnected = false;
	}

	~Server(){
		if (isConnected) closeConnection();
	}

	static bool setupServer(int port){
		if ((serverSocket = socket(AF_INET, SOCK_STREAM, 0)) == -1) return false;
		
		int off = 1;
		setsockopt(serverSocket, IPPROTO_TCP, TCP_NODELAY, (char *) &off, sizeof(int));

		server.sin_family      = AF_INET;         
		server.sin_port        = htons(port);
		server.sin_addr.s_addr = INADDR_ANY;
		memset(&server.sin_zero, 0, 8);
		
		if (bind(serverSocket, (sockaddr *) &server, sizeof(server)) == -1) return false;
		if (listen(serverSocket, 2) == -1) return false;
		return true;
	}

	bool acceptConnection(){
		int size = sizeof(client);
		if ((clientSocket = accept(serverSocket, (sockaddr *) &client, &size)) == -1) return false;
		
		int off = 1;
		setsockopt(clientSocket, IPPROTO_TCP, TCP_NODELAY, (char *) &off, sizeof(int));

		//printf("You got a connection from %s\n", inet_ntoa(client.sin_addr));
		isConnected = true;
		return true;		
	}

	bool closeConnection(){
		isConnected = false;
		return (closesocket(clientSocket) != SOCKET_ERROR);
	}

	bool sendMessage(void *buffer, int length){
		return (send(clientSocket, (char *) buffer, length, 0) != SOCKET_ERROR);
	}

	int getMessage(void *buffer, int length){
		return recv(clientSocket, (char *) buffer, length, 0);
	}

};

SOCKET Server::serverSocket;
sockaddr_in Server::server;

/* ------------------------------------- */

class Client {
protected:
	SOCKET serverSocket;
	sockaddr_in server;
	bool isConnected;
public:
	Client(){
		isConnected = false;
	}

	~Client(){
		if (isConnected) disconnect();
	}

	bool connectTo(char *address, int port){
		if ((serverSocket = socket(AF_INET, SOCK_STREAM, 0)) == -1) return false;
		
		int off = 1;
		setsockopt(serverSocket, IPPROTO_TCP, TCP_NODELAY, (char *) &off, sizeof(int));

		hostent *he;
		if ((he = gethostbyname(address)) == NULL) return false;		
		
		server.sin_family = AF_INET;
		server.sin_port   = htons(port);
		server.sin_addr   = *((in_addr *) he->h_addr);
		memset(&server.sin_zero, 0, 8);
		
		if (connect(serverSocket, (sockaddr *) &server, sizeof(sockaddr)) == -1) return false;
		isConnected = true;

		return true;
	}

	bool disconnect(){
		isConnected = false;
		return (closesocket(serverSocket) != SOCKET_ERROR);
	}

	bool sendMessage(void *buffer, int length){
		return (send(serverSocket, (char *) buffer, length, 0) != SOCKET_ERROR);
	}

	int getMessage(void *buffer, int length){
		return recv(serverSocket, (char *) buffer, length, 0);
	}


};


#endif //_NETWORK_H_