#ifndef _TOKENIZER_H_
#define _TOKENIZER_H_

class Tokenizer {
private:
	char *str;
	unsigned int length;
	unsigned int start, end;

public:
	static bool isDelimiter(char ch){
		return (ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n');
	}

	static bool isNumeric(char ch){
		return (ch >= '0' && ch <= '9');
	}

	static bool isAlphabetical(char ch){
		return ((ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z') || ch == '_');
	}

	Tokenizer();
	~Tokenizer();

	void setString(char *string);
	bool setFile(char *fileName);
	void reset();
	bool next(char *buffer);
	bool nextAfterToken(char *buffer, char *token);

};

/* ------------------------------------------------------------------- */

Tokenizer::Tokenizer(){
	str = NULL;

	reset();
}

Tokenizer::~Tokenizer(){
	if (str != NULL) delete str;
}

void Tokenizer::setString(char *string){
	if (str != NULL) delete str;

	length = strlen(string);

	str = new char[(length + 1) * sizeof(char)];
	strcpy(str, string);

	reset();
}

bool Tokenizer::setFile(char *fileName){
	FILE *file;

	if (str != NULL) delete str;

	if ((file = fopen(fileName, "rb")) != NULL){
		fseek(file, 0, SEEK_END);
		length = ftell(file);
		fseek(file, 0, SEEK_SET);

		str = new char[(length + 1) * sizeof(char)];
		fread(str, length, 1, file);
		str[length] = '\0';

		fclose(file);
		
		reset();
	}
	return false;
}

void Tokenizer::reset(){
	end = 0;
}

bool Tokenizer::next(char *buffer){
	start = end;

	while (start < length && isDelimiter(str[start])) start++;
	if (start < length){
		
		end = start + 1;
		
		if (isNumeric(str[start])){
			while (isNumeric(str[end]) || str[end] == '.') end++;
		} else if (isAlphabetical(str[start])){
			while (isAlphabetical(str[end]) || isNumeric(str[end])) end++;
		}

		if (buffer != NULL){
			strncpy(buffer, str + start, end - start);
			buffer[end - start] = '\0';
		}
		return true;
	}
	return false;
}

bool Tokenizer::nextAfterToken(char *buffer, char *token){
	char str[256];

	while (next(str)){
		if (stricmp(str, token) == 0){
			return next(buffer);
		}
	}

	return false;
}


#endif // _TOKENIZER_H_