                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--


  In accordance with the second point I'd like to credit the authors of some
  parts of the code in here:

  TODO:
   * Fix the list of authors ...



*/

#ifndef _IMAGE_H_
#define _IMAGE_H_

#define IMAGE_SUCCESSFUL                0
#define IMAGE_FILE_NOT_FOUND            1
#define IMAGE_FILETYPE_NOT_SUPPORTED    2
#define IMAGE_INTERNAL_ERROR            3
#define IMAGE_IMPROPER_OPERAND          4


#define IMAGE_FORMAT_UNCOMPRESSED       0
#define IMAGE_FORMAT_COMPRESSED_DXT1    1
#define IMAGE_FORMAT_COMPRESSED_DXT3    2
#define IMAGE_FORMAT_COMPRESSED_DXT5    3

#define IMAGE_FORMAT_MIPMAPPED 0x80000000

#include "ImageUtils.h"

// ##############################################################################################################################

unsigned int loadBmp(const char *File, unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels, unsigned int &format){
	HANDLE FileHandle;
	BY_HANDLE_FILE_INFORMATION info;
	unsigned int FileSize,pos,x,y,i,j,col;
	unsigned char *Data;

	FileHandle = CreateFile(File,GENERIC_READ,0,NULL,OPEN_EXISTING,0,NULL);
	if (FileHandle == INVALID_HANDLE_VALUE)	return IMAGE_FILE_NOT_FOUND;
	GetFileInformationByHandle(FileHandle,&info);

	Data = (unsigned char *) malloc(info.nFileSizeLow);
	
	ReadFile(FileHandle,Data,info.nFileSizeLow,(unsigned long *) &FileSize,NULL);
	CloseHandle(FileHandle);

	int compression = ((unsigned __int16 *) Data)[15];

	width  = ((unsigned __int16 *) Data)[9];
	height = ((unsigned __int16 *) Data)[11];
	channels = 3;
	format = IMAGE_FORMAT_UNCOMPRESSED;

	switch (((unsigned __int16 *) Data)[14]){
	case 8:
		pixels = (unsigned char *) malloc(3 * width * height);

		if (compression == 0){
			// Standard 8 bit palettized bitmap
			pos = 0;
			i = 3 * width * (height - 1);
			for (y = 0; y < height; y++){
				for (x = 0; x < width; x++){
					col = Data[pos + 1078];
					pixels[i + 3 * x    ] = Data[56 + (col << 2)];
					pixels[i + 3 * x + 1] = Data[55 + (col << 2)];
					pixels[i + 3 * x + 2] = Data[54 + (col << 2)];
					pos++;
				}
				i -= 3 * width;
				pos = (pos + 3) & ~3;
			}
		} else {
			// RLE encoded 8 bit palettized bitmap
			pos = 1078;
			unsigned int t = 0,n,i;
			bool endOfBitmap = false;

			while (!endOfBitmap){
				n   = Data[pos];
				col = Data[pos+1];
				if (n != 0){

					for (i = 0; i < n; i++){
						pixels[t  ] = Data[56 + (col << 2)];
						pixels[t+1] = Data[55 + (col << 2)];
						pixels[t+2] = Data[54 + (col << 2)];
						t += 3;
					}
					pos += 2;
				} else {
					if (col < 2){
						endOfBitmap = (col == 1);
						pos += 2;
					} else {
						n = col;
						
						pos += 2;
						for (i = 0; i < n; i++){
							col = Data[pos];
							pos++;
							pixels[t  ] = Data[56 + (col << 2)];
							pixels[t+1] = Data[55 + (col << 2)];
							pixels[t+2] = Data[54 + (col << 2)];
							t += 3;
						}
						if (pos & 1) pos++;
					}
				}
			}
			// Flip vertically
			unsigned char tmp;
			i = 0;
			j = 3 * width * (height - 1);
			for (y = 0; y < height >> 1; y++){
				for (x = 0; x < 3 * width; x++){
					tmp = pixels[i + x];
					pixels[i + x] = pixels[j + x];
					pixels[j + x] = tmp;
				}
				i += 3 * width;
				j -= 3 * width;
			}
		}
		break;
	case 24:
			// Standard 24 bit RGB bitmap
		pixels = (unsigned char *) malloc(3 * width * height);
		pos = 0;

		i = 3 * width * (height - 1);
		for (y = 0; y < height; y++){
			for (x = 0; x < width; x++){
				pixels[i + 3 * x    ] = Data[56 + pos];
				pixels[i + 3 * x + 1] = Data[55 + pos];
				pixels[i + 3 * x + 2] = Data[54 + pos];
				pos += 3;
			}
			i -= 3 * width;
			pos = (pos + 3) & ~3;
		}
		break;
	default:
		free(Data);
		return IMAGE_FILETYPE_NOT_SUPPORTED;
	}

	free(Data);

	return IMAGE_SUCCESSFUL;
}



// ##############################################################################################################################



unsigned int loadTarga(const char *File, unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels, unsigned int &format){
	#pragma pack (push)
	#pragma pack (1)
	struct Tgaheader {
		__int8	descriptionlen;
		__int8	cmaptype;
		__int8	imagetype;
		__int16	cmapstart;
		__int16	cmapentries;
		__int8	cmapbits;
		__int16	xoffset;
		__int16	yoffset;
		__int16	width;
		__int16	height;
		__int8	bpp;
		__int8	attrib;
	};
	#pragma pack (pop)

	Tgaheader header;
	unsigned int size,x,y,dataChannels;
	unsigned char *Data, *fBuffer;
	unsigned int fPointer;
	unsigned int tempPixel;
	unsigned char Palette[768];
	HANDLE FileHandle;
	BY_HANDLE_FILE_INFORMATION info;
	unsigned int FileSize;
	

	FileHandle = CreateFile(File,GENERIC_READ,0,NULL,OPEN_EXISTING,0,NULL);
	if (FileHandle == INVALID_HANDLE_VALUE)	return IMAGE_FILE_NOT_FOUND;
	GetFileInformationByHandle(FileHandle,&info);

	ReadFile(FileHandle,&header,sizeof(header),(unsigned long *) &FileSize,NULL);
	
	width = header.width;
	height = header.height;
	format = IMAGE_FORMAT_UNCOMPRESSED;

	switch (header.bpp){
	case 8:
		channels = 3;
		dataChannels = 1;
		break;
	case 16:
		channels = 4;
		dataChannels = 2;
		break;
	case 24:
		channels = 3;
		dataChannels = 3;
		break;
	case 32:
		channels = 4;
		dataChannels = 4;
		break;
	default:
		CloseHandle(FileHandle);
		return IMAGE_FILETYPE_NOT_SUPPORTED;
	}

	int palLength = header.descriptionlen + header.cmapentries * header.cmapbits / 8;
	if (palLength > 0)
		ReadFile(FileHandle,&Palette,palLength,(unsigned long *) &FileSize,NULL);
	
	fBuffer = (unsigned char *) malloc(info.nFileSizeLow - sizeof(header) - palLength);
	
	ReadFile(FileHandle,fBuffer,info.nFileSizeLow - sizeof(header) - palLength,(unsigned long *) &FileSize,NULL);
	CloseHandle(FileHandle);


	size = width * height * dataChannels;
	fPointer = 0;


	pixels = (unsigned char *) malloc(width * height * channels);

	Data = (unsigned char *) malloc(size);

	
	// Image can be raw or rle, bit 3 of .imagetype tells.
	if (header.imagetype & 0x08){
		unsigned char v[16];		// temporary array to store pixel components
		unsigned char *buffer = Data;
		unsigned int c,count;
		
		while (size > 0) {
			// get packet header
			c = fBuffer[fPointer];
			fPointer++;

			count = (c & 0x7f) + 1;
			
			size -= count * dataChannels;
			
			if (c & 0x80) {
				// rle packet
				memcpy(v,fBuffer + fPointer,dataChannels);// load pixel components
				fPointer += dataChannels;

				while (count > 0){
					memcpy(buffer,v,dataChannels);
					buffer += dataChannels;
					count--;
				}
			} else {
				// raw packet
				count *= dataChannels;
				memcpy(buffer,fBuffer + fPointer,count);
				fPointer += count;				

				buffer += count;
			}
		}
		
	} else {
		memcpy(Data,fBuffer,size);
	}
	
	switch(header.bpp) {
	case 8:
		for (y = 0; y < height; y++){
			for (x = 0; x < width; x++){
				tempPixel = Data[(width * (height - y - 1) + x)];
				pixels[3 * (y * width + x)    ] = Palette[3 * tempPixel + 2];
				pixels[3 * (y * width + x) + 1] = Palette[3 * tempPixel + 1];
				pixels[3 * (y * width + x) + 2] = Palette[3 * tempPixel    ];
			}
		}
		break;
	case 16:
		for (y = 0; y < height; y++){
			for (x = 0; x < width; x++){
				__int16 temp = ((__int16 *) Data)[(width * (height - y - 1) + x)];

				pixels[4 * (y * width + x)    ] = ((temp >> 10) & 0x1F) << 3;
				pixels[4 * (y * width + x) + 1] = ((temp >> 5) & 0x1F) << 3;
				pixels[4 * (y * width + x) + 2] = (temp & 0x1F) << 3;
				pixels[4 * (y * width + x) + 3] = (temp >> 15)? 255:0;
			}
		}
		break;
	case 24:
		for (y = 0; y < height; y++){
			for (x = 0; x < width; x++){
				pixels[3 * (y * width + x)    ] = Data[3 * (width * (height - y - 1) + x) + 2];
				pixels[3 * (y * width + x) + 1] = Data[3 * (width * (height - y - 1) + x) + 1];
				pixels[3 * (y * width + x) + 2] = Data[3 * (width * (height - y - 1) + x)    ];
			}
		}
		break;
	case 32:
		for (y = 0; y < height; y++){
			for (x = 0; x < width; x++){
				pixels[4 * (y * width + x)    ] = Data[4 * (width * (height - y - 1) + x) + 2];
				pixels[4 * (y * width + x) + 1] = Data[4 * (width * (height - y - 1) + x) + 1];
				pixels[4 * (y * width + x) + 2] = Data[4 * (width * (height - y - 1) + x)    ];
				pixels[4 * (y * width + x) + 3] = Data[4 * (width * (height - y - 1) + x) + 3];
			}
		}
		break;
	}

	free(Data);
	free(fBuffer);

	return IMAGE_SUCCESSFUL;
}



// ##############################################################################################################################


unsigned int loadPcx(const char *File, unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels, unsigned int &format){
	HANDLE FileHandle;
	BY_HANDLE_FILE_INFORMATION info;
	unsigned int FileSize,t,x,y,pos,col,length,size,bpp;
	unsigned char *Data;

	FileHandle = CreateFile(File,GENERIC_READ,0,NULL,OPEN_EXISTING,0,NULL);
	if (FileHandle == INVALID_HANDLE_VALUE)	return IMAGE_FILE_NOT_FOUND;
	GetFileInformationByHandle(FileHandle,&info);

	Data = (unsigned char *) malloc(info.nFileSizeLow);
	
	ReadFile(FileHandle,Data,info.nFileSizeLow,(unsigned long *) &FileSize,NULL);
	CloseHandle(FileHandle);

	width  = 1 + *((unsigned __int16 *) (Data + 8));
	height = 1 + *((unsigned __int16 *) (Data + 10));
	bpp    = Data[65] * Data[3];
	channels = 3;
	format = IMAGE_FORMAT_UNCOMPRESSED;

	pixels = (unsigned char *) malloc(width * height * channels);

	switch(bpp){
	case 8:
		t = 128;
		pos = 0;
		while (pos < width * height * channels){
			if (Data[t] < 192){
				col = Data[t];
				pixels[pos  ] = Data[FileSize - 768 + 3 * col];
				pixels[pos+1] = Data[FileSize - 767 + 3 * col];
				pixels[pos+2] = Data[FileSize - 766 + 3 * col];
				pos += 3;
				t++;
			} else {
				length = Data[t] - 192;
				col = Data[t+1];
				for (x = 0; x < length; x++){
					pixels[pos  ] = Data[FileSize - 768 + 3 * col];
					pixels[pos+1] = Data[FileSize - 767 + 3 * col];
					pixels[pos+2] = Data[FileSize - 766 + 3 * col];
					pos += 3;
				}
				t += 2;
			}
		}
		break;
	case 24:
		t = 128;
		pos = 0;
		size = width * height * channels;
		while (pos < size){
			if (Data[t] < 192){
				pixels[pos++] = Data[t++];
			} else {
				length = Data[t++] - 192;
				for (x = 0; x < length; x++){
					pixels[pos++] = Data[t];
				}
				t++;
			}
		}
		
		// RRR RRR ... GGG GGG... BBB BBB... =>  RGB RGB RGB RGB RGB RGB ...
		
		unsigned char *dest = pixels;
		unsigned char *temp = (unsigned char *) malloc(width * 3);
		for (y = 0; y < height; y++){
			memcpy(temp, dest, width * 3);
			for (x = 0; x < width; x++){
				dest[3 * x    ] = temp[x];
				dest[3 * x + 1] = temp[x + width];
				dest[3 * x + 2] = temp[x + width * 2];
			}
			dest += width * 3;
		}
		free(temp);
		break;
	}

	free(Data);

	return IMAGE_SUCCESSFUL;
}

// ##############################################################################################################################

#ifndef NO_PNG

#include "png.h"
#include "cexcept.h"

define_exception_type(const char *);
extern struct exception_context the_exception_context[1];
struct exception_context the_exception_context[1];


unsigned int loadPng(const char *File, unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels, unsigned int &format){
	static png_structp png_ptr = NULL;
	static png_infop info_ptr = NULL;
    static FILE *pfFile;
	png_const_charp msg;

    png_byte pbSig[8];
    int iBitDepth, iColorType;
    double dGamma;
    png_color_16 *pBackground;
    png_uint_32 ulChannels;
    png_uint_32 ulRowBytes;
    png_byte *pbImageData = pixels;
    static png_byte **ppbRowPointers = NULL;
	
    // open the PNG input file
    if (!(pfFile = fopen(File, "rb"))){
        pixels = pbImageData = NULL;
        return IMAGE_FILE_NOT_FOUND;
    }
	
    // first check the eight byte PNG signature
    fread(pbSig, 1, 8, pfFile);
    if (!png_check_sig(pbSig, 8)){
        pixels = pbImageData = NULL;
        return IMAGE_FILETYPE_NOT_SUPPORTED;
    }
	
    // create the two png(-info) structures
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, (png_error_ptr) NULL, (png_error_ptr) NULL);
    if (!png_ptr){
        pixels = pbImageData = NULL;
        return IMAGE_INTERNAL_ERROR;
    }
	
    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr){
        png_destroy_read_struct(&png_ptr, NULL, NULL);
        pixels = pbImageData = NULL;
        return IMAGE_INTERNAL_ERROR;
    }
	
    Try {
        // initialize the png structure
        png_init_io(png_ptr, pfFile);
        png_set_sig_bytes(png_ptr, 8);
        
        // read all PNG info up to image data
        png_read_info(png_ptr, info_ptr);
        
        // get width, height, bit-depth and color-type
        png_get_IHDR(png_ptr, info_ptr, (unsigned long *) &width, (unsigned long *) &height, &iBitDepth,
            &iColorType, NULL, NULL, NULL);
        
        // expand images of all color-type and bit-depth to 3x8 bit RGB images
        // let the library process things like alpha, transparency, background
        if (iBitDepth == 16)
            png_set_strip_16(png_ptr);
        if (iColorType == PNG_COLOR_TYPE_PALETTE)
            png_set_expand(png_ptr);
        if (iBitDepth < 8)
            png_set_expand(png_ptr);
        if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
            png_set_expand(png_ptr);
        if (iColorType == PNG_COLOR_TYPE_GRAY ||
            iColorType == PNG_COLOR_TYPE_GRAY_ALPHA)
            png_set_gray_to_rgb(png_ptr);
        
        // set the background color to draw transparent and alpha images over.
        if (png_get_bKGD(png_ptr, info_ptr, &pBackground)){
            png_set_background(png_ptr, pBackground, PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
        }
        
        // if required set gamma conversion
        if (png_get_gAMA(png_ptr, info_ptr, &dGamma))
            png_set_gamma(png_ptr, (double) 2.2, dGamma);
        
        // after the transformations have been registered update info_ptr data
        png_read_update_info(png_ptr, info_ptr);
        
        // get again width, height and the new bit-depth and color-type
        png_get_IHDR(png_ptr, info_ptr, (unsigned long *) &width, (unsigned long *) &height, &iBitDepth, &iColorType, NULL, NULL, NULL);
        
        
        // row_bytes is the width x number of channels
        ulRowBytes = png_get_rowbytes(png_ptr, info_ptr);
        ulChannels = png_get_channels(png_ptr, info_ptr);
        
        channels = ulChannels;
        
        // now we can allocate memory to store the image
        if ((pbImageData = (png_byte *) malloc(ulRowBytes * height * sizeof(png_byte))) == NULL){
            png_error(png_ptr, "Visual PNG: out of memory");
        }
        pixels = pbImageData;
        
        // and allocate memory for an array of row-pointers
        if ((ppbRowPointers = (png_bytepp) malloc(height * sizeof(png_bytep))) == NULL){
            png_error(png_ptr, "Visual PNG: out of memory");
        }
        
        // set the individual row-pointers to point at the correct offsets
        for (unsigned int i = 0; i < height; i++)
            ppbRowPointers[i] = pbImageData + i * ulRowBytes;
        
        // now we can go ahead and just read the whole image
        png_read_image(png_ptr, ppbRowPointers);
        
        // read the additional chunks in the PNG file (not really needed)
        png_read_end(png_ptr, NULL);
        
        // and we're done
        free(ppbRowPointers);
        ppbRowPointers = NULL;
        
        // yepp, done

    } Catch (msg) {
        png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
		
        pixels = pbImageData = NULL;
        
        if (ppbRowPointers)
           free(ppbRowPointers);
		
        fclose(pfFile);
		
        return IMAGE_INTERNAL_ERROR;
    }
	
    png_destroy_read_struct(&png_ptr, &info_ptr, NULL);

    fclose (pfFile);

	format = IMAGE_FORMAT_UNCOMPRESSED;
	
    return IMAGE_SUCCESSFUL;
}

#endif // NO_PNG


// ##############################################################################################################################


#ifndef NO_JPEG

extern "C" {
#include "jpeglib.h"
}


unsigned int loadJpeg(const char *fileName, unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels, unsigned int &format){
	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;

	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_decompress(&cinfo);

	FILE *infile;
	if ((infile = fopen(fileName, "rb")) == NULL){
	    return IMAGE_FILE_NOT_FOUND;
	}
	jpeg_stdio_src(&cinfo, infile);
	jpeg_read_header(&cinfo, TRUE);
	jpeg_start_decompress(&cinfo);

	channels = cinfo.num_components;
	width    = cinfo.output_width;
	height   = cinfo.output_height;
	format   = IMAGE_FORMAT_UNCOMPRESSED;

	pixels = (unsigned char *) malloc(channels * width * height);
	unsigned char *curr_scanline = pixels;

	while (cinfo.output_scanline < height){
		jpeg_read_scanlines(&cinfo, &curr_scanline, 1);
		curr_scanline += channels * width;
	}
	

	jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);

	fclose(infile);

	return IMAGE_SUCCESSFUL;
}

#endif // NO_JPEG


// ##############################################################################################################################

#ifndef NO_GIF

extern "C" {
#include "GifMalloc.h"
#include "RGB.h"
#include "Gif.h"
};


unsigned int loadGif(const char *fileName, unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels, unsigned int &format){
	Gif *gif = read_gif_file((char *) fileName);

	// No support for animated gifs right now ...
	if (gif->block_count == 1){

		width  = gif->screen->width;
		height = gif->screen->height;
		channels = 3;
		format = IMAGE_FORMAT_UNCOMPRESSED;
		
		pixels = (unsigned char *) malloc(width * height * channels);
		
		Colour col;
		
		unsigned int x,y;
		unsigned char *scanline, *dest = pixels;
		
		for (y = 0; y < height; y++){
			scanline = gif->blocks[0]->pic->data[y];
			for (x = 0; x < width; x++){
				col = gif->screen->cmap->colours[scanline[x]];
				
				*dest++ = col.red;
				*dest++ = col.green;
				*dest++ = col.blue;
			}
		}
		
		del_gif(gif);
		return IMAGE_SUCCESSFUL;

	} else {

		del_gif(gif);
		return IMAGE_FILETYPE_NOT_SUPPORTED;
	}
}

#endif // NO_GIF

// ##############################################################################################################################

#ifndef NO_DDS

// DDS loading code requires that DirectX SDK is installed to compile due to the DDSURFACEDESC2 struct header
#include <ddraw.h>

unsigned int loadDDS(const char *fileName, unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels, unsigned int &format){
	DDSURFACEDESC2 ddsd;
	
	FILE *fp = fopen(fileName, "rb");
	if (fp == NULL) return IMAGE_FILE_NOT_FOUND;

	char filecode[4];
	fread(filecode, 1, 4, fp);
	if (strncmp(filecode, "DDS ", 4) != 0){
		fclose(fp);
		return IMAGE_FILETYPE_NOT_SUPPORTED;
	}

	fread(&ddsd, sizeof(ddsd), 1, fp);

	width    =  ddsd.dwWidth;
	height   =  ddsd.dwHeight;
	channels = (ddsd.ddpfPixelFormat.dwFourCC == FOURCC_DXT1) ? 3 : 4;

	switch(ddsd.ddpfPixelFormat.dwFourCC){
	case FOURCC_DXT1:
		format = IMAGE_FORMAT_COMPRESSED_DXT1;
		break;
	case FOURCC_DXT3:
		format = IMAGE_FORMAT_COMPRESSED_DXT3;
		break;
	case FOURCC_DXT5:
		format = IMAGE_FORMAT_COMPRESSED_DXT5;
		break;
	default:
		return IMAGE_FILETYPE_NOT_SUPPORTED;
	}
	
	
	// Calculate how large buffer we need possibly including mipmaps
	int size, bufsize = 0, w = width, h = height;

	if (ddsd.dwMipMapCount > 1){
        unsigned int bytesPerBlock = (format == IMAGE_FORMAT_COMPRESSED_DXT1)? 8 : 16;
		while (w > 1 || h > 1){
			//size = w * h;
            size = ((w + 3) >> 2) * ((h + 3) >> 2) * bytesPerBlock;

			bufsize += size;//(size < 16)? 16 : size;

			if (w > 1) w >>= 1;
			if (h > 1) h >>= 1;
		}
		bufsize += bytesPerBlock;

		format |= IMAGE_FORMAT_MIPMAPPED;
	} else {
		bufsize = width * height;
	}

	/*if (format == IMAGE_FORMAT_COMPRESSED_DXT1){
		bufsize >>= 1;
	}*/

	pixels = (unsigned char *) malloc(bufsize * sizeof(unsigned char));
	fread(pixels, 1, bufsize, fp);
	fclose(fp);

	return IMAGE_SUCCESSFUL;
}


#endif // NO_DDS

// ##############################################################################################################################

unsigned int getMipMappedImageSize(unsigned int width, unsigned int height, unsigned int channels){
	unsigned int size = 1;

	while (height > 1 || width > 1){
		size += width * height;
		
		if (width  > 1) width  >>= 1;
		if (height > 1) height >>= 1;

	}

	return size * channels;
}

void uncompressDXT(unsigned char *dest, unsigned char *&src, unsigned int width, unsigned int height, unsigned channels, unsigned int format){
	unsigned char *dst;
	unsigned int x,y,i,j,k, c0,c1, indexes, index, sx,sy;
	unsigned char colors[4][3];
	unsigned __int64 alpha;
	unsigned char a0, a1;

	sx = (width  < 4)? width  : 4;
	sy = (height < 4)? height : 4;
	
	for (y = 0; y < height; y += 4){
		for (x = 0; x < width; x += 4){
			if (format == IMAGE_FORMAT_COMPRESSED_DXT3){
				alpha = *(unsigned __int64 *) src;
				src += 8;
			} else if (format == IMAGE_FORMAT_COMPRESSED_DXT5){
				alpha = (*(unsigned __int64 *) src) >> 16;
				a0 = src[0];
				a1 = src[1];
				src += 8;
			}
			c0 = *(unsigned __int16 *) src;
			c1 = *(unsigned __int16 *) (src + 2);
			
			colors[0][0] = ((c0 >> 11) & 0x1F) << 3;
			colors[0][1] = ((c0 >>  5) & 0x3F) << 2;
			colors[0][2] =  (c0        & 0x1F) << 3;
			
			colors[1][0] = ((c1 >> 11) & 0x1F) << 3;
			colors[1][1] = ((c1 >>  5) & 0x3F) << 2;
			colors[1][2] =  (c1        & 0x1F) << 3;

			if (c0 > c1 || format == IMAGE_FORMAT_COMPRESSED_DXT5){
				for (i = 0; i < 3; i++){
					colors[2][i] = (2 * colors[0][i] +     colors[1][i] + 1) / 3;
					colors[3][i] = (    colors[0][i] + 2 * colors[1][i] + 1) / 3;
				}
			} else {
				for (i = 0; i < 3; i++){
					colors[2][i] = (colors[0][i] + colors[1][i] + 1) >> 1;
					colors[3][i] = 0;
				}
			}
			
			src += 4;
			for (i = 0; i < sy; i++){
				dst = dest + ((y + i) * width + x) * channels;
				indexes = src[i];
				for (j = 0; j < sx; j++){
					index = indexes & 0x3;
					for (k = 0; k < 3; k++){
						*dst++ = colors[index][k];
					}
					if (channels == 4){
						if (format == IMAGE_FORMAT_COMPRESSED_DXT1){
							*dst++ = (index == 3 && c0 <= c1)? 0 : 255;
						} else if (format == IMAGE_FORMAT_COMPRESSED_DXT3){
							*dst++ = (unsigned char) (alpha & 0xF) << 4;
							alpha >>= 4;
						} else if (format == IMAGE_FORMAT_COMPRESSED_DXT5){
							k = ((unsigned int) alpha) & 0x7;
							if (k == 0){
								*dst++ = a0;
							} else if (k == 1){
								*dst++ = a1;
							} else if (a0 > a1){
								*dst++ = ((8 - k) * a0 + (k - 1) * a1) / 7;
							} else if (k >= 6){
								*dst++ = (k == 6)? 0 : 255;
							} else {
								*dst++ = ((6 - k) * a0 + (k - 1) * a1) / 5;
							}
							alpha >>= 3;
						}
					}
					indexes >>= 2;
				}
			}
			src += 4;
		}
	}
}

unsigned int uncompressImage(unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels, unsigned int &format){
	unsigned char *src, *dest, *newPixels = (unsigned char *) malloc(getMipMappedImageSize(width, height, channels));
	unsigned int w,h;
	unsigned int mipmapped = format &  IMAGE_FORMAT_MIPMAPPED;
	unsigned int imformat  = format & ~IMAGE_FORMAT_MIPMAPPED;

	src = pixels;
	w = 2 * width;
	h = 2 * height;
	dest = newPixels;

	do {
		if (w > 1) w >>= 1;
		if (h > 1) h >>= 1;

		switch(imformat){
		case IMAGE_FORMAT_COMPRESSED_DXT1:
		case IMAGE_FORMAT_COMPRESSED_DXT3:
		case IMAGE_FORMAT_COMPRESSED_DXT5:
			uncompressDXT(dest, src, w, h, channels, imformat);
			break;
		}
	
		dest += w * h * channels;
	} while (mipmapped && (w > 1 || h > 1));

	format = IMAGE_FORMAT_UNCOMPRESSED | mipmapped;

	free(pixels);
	pixels = newPixels;
	return IMAGE_SUCCESSFUL;
}


// ##############################################################################################################################

/* 
	* Generic image loader
*/
unsigned int loadImage(const char *File, unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels, unsigned int &format){
	char *extension;
	
	extension = strrchr(File,'.');

	if (extension != NULL){
		if (stricmp(extension,".tga") == 0){
			return loadTarga(File, pixels, width, height, channels, format);
		} else if (stricmp(extension,".pcx") == 0){
			return loadPcx(File, pixels, width, height, channels, format);
		} else if (stricmp(extension,".bmp") == 0){
			return loadBmp(File, pixels, width, height, channels, format);
		} 
#ifndef NO_JPEG
		else if (stricmp(extension,".jpg") == 0 || stricmp(extension,".jpeg") == 0){
			return loadJpeg(File, pixels, width, height, channels, format);
		} 
#endif
#ifndef NO_PNG
		else if (stricmp(extension,".png") == 0){
			return loadPng(File, pixels, width, height, channels, format);
		} 
#endif
#ifndef NO_GIF
		else if (stricmp(extension,".gif") == 0){
			return loadGif(File, pixels, width, height, channels, format);
		} 
#endif
#ifndef NO_DDS
		else if (stricmp(extension,".dds") == 0){
			return loadDDS(File, pixels, width, height, channels, format);
		} 
#endif
		else return IMAGE_FILETYPE_NOT_SUPPORTED;
	}
	return IMAGE_FILETYPE_NOT_SUPPORTED;
}



#endif // _IMAGE_H_
