                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/


#ifndef _SHADER_H_
#define _SHADER_H_

typedef int ShaderID;

#define SHADER_DISABLED (-1)

/* ################################################################################################################### */

class VertexShader {
protected:
    GLuint shaderID;
    GLuint firstInvariantsID;
public:
    VertexShader(){}
    ~VertexShader(){}

    void beginVertexShader(){
        shaderID = glGenVertexShadersEXT(1);
        glBindVertexShaderEXT(shaderID);
        glBeginVertexShaderEXT();
    }

    void endVertexShader(GLuint firstInvariant){
        firstInvariantsID = firstInvariant;
        glEndVertexShaderEXT();
    }

    GLuint getInvariants(){
        return firstInvariantsID;
    }

    bool activate(){
        glBindVertexShaderEXT(shaderID);
        return true;
    }

    bool deactivate(){
        return true;
    }
};


/* ################################################################################################################### */

class GeneralFragmentShader {
protected:
    GLuint shaderID;
public:
    GeneralFragmentShader(){}
    virtual ~GeneralFragmentShader(){}

    virtual GLint getFragmentShaderType() = 0;

    virtual void beginFragmentShader() = 0;
    virtual void endFragmentShader() = 0;

    virtual bool activate() = 0;
    virtual bool deactivate() = 0;
};

/* ----------------------------------------------------*/

class FragmentShader : public GeneralFragmentShader {
public:
    FragmentShader(){
    }
    ~FragmentShader(){
        glDeleteFragmentShaderATI(shaderID);
    }

    GLint getFragmentShaderType(){
        return GL_FRAGMENT_SHADER_ATI;
    }

    void beginFragmentShader(){
        shaderID = glGenFragmentShadersATI(1);
	    glBindFragmentShaderATI(shaderID);
    	glBeginFragmentShaderATI();
    }

    void endFragmentShader(){
    	glEndFragmentShaderATI();
    }

    bool activate(){
    	glBindFragmentShaderATI(shaderID);
        return true;
    }

    bool deactivate(){
        return true;
    }
};

/* ----------------------------------------------------*/

class RegisterCombiner : public GeneralFragmentShader {
public:
    RegisterCombiner(){
    }
    ~RegisterCombiner(){
        glDeleteLists(shaderID, 1);
    }

    GLint getFragmentShaderType(){
        return GL_REGISTER_COMBINERS_NV;
    }

    void beginFragmentShader(){
        shaderID = glGenLists(1);
        glNewList(shaderID, GL_COMPILE);
    }

    void endFragmentShader(){
    	glEndList();
    }

    bool activate(){
    	glCallList(shaderID);
        return true;
    }

    bool deactivate(){
        return true;
    }
};



/* ################################################################################################################### */


#endif // _SHADER_H_