                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/


#ifndef _TEXTURE_H_
#define _TEXTURE_H_

typedef int TextureID;

#define TEXTURE_DISABLED (-1)
#define TEXTURE_INVALID  (-2)

enum TexEnv {
	TEXENV_INVALID,
	TEXENV_MODULATE,
    TEXENV_MODULATE_PRIMARY,
	TEXENV_MODULATE_NEXT,
	TEXENV_MODULATE_PREV_NEXT,
	TEXENV_REPLACE,
	TEXENV_ADD,
	TEXENV_ADD_PRIMARY,
    TEXENV_INTERPOLATE,
	TEXENV_DOT3,
	TEXENV_DOT3_NEXT,
};

#define TEXFLAGS_RENORMALIZE  0x1

#define BLEND_INVALID (-1)

#include "../Imaging/Image.h"
#include "../3D/TextureUtils.h"
#include "Shader.h"

/* ################################################################################################################### */

class Texture {
protected:
	GLuint textureID;
	GLenum target;

	GLint minFilter, magFilter;
	GLint wrapModeS, wrapModeT, wrapModeR;
	GLint internalFormat;

	unsigned int width, height, depth, channels;

public:
	Texture();
	virtual ~Texture();

	GLenum getTarget(){
		return target;
	}

	void setFilter(GLint filter);
	void setWrapMode(GLint wrapS, GLint wrapT, GLint wrapR);
    void setInternalFormat(GLint iFormat);
	virtual bool upload() = 0;
	virtual bool unload();
	virtual bool activate();
	virtual bool deactivate();
};

/* ----------------------------------------------------*/

Texture::Texture(){
	width = height = depth = channels = 0;
	textureID = TEXTURE_INVALID;
	minFilter = GL_LINEAR_MIPMAP_LINEAR;
	magFilter = GL_LINEAR;
	wrapModeS = GL_REPEAT;
	wrapModeT = GL_REPEAT;
	wrapModeR = GL_REPEAT;
	internalFormat = GL_RGB8;

	//target = ??
}

Texture::~Texture(){

}

void Texture::setFilter(GLint filter){
	minFilter = filter;
	if (filter == GL_NEAREST || filter == GL_NEAREST_MIPMAP_LINEAR){
		magFilter = GL_NEAREST;
	} else {
		magFilter = GL_LINEAR;
	}
}

void Texture::setWrapMode(GLint wrapS, GLint wrapT, GLint wrapR = GL_REPEAT){
	wrapModeS = wrapS;
	wrapModeT = wrapT;
    wrapModeR = wrapR;
}

void Texture::setInternalFormat(GLint iFormat){
    internalFormat = iFormat;
}

bool Texture::unload(){
	if (textureID != TEXTURE_INVALID){
		glDeleteTextures(1,&textureID);
		textureID = TEXTURE_INVALID;
	}
	return true;
}

bool Texture::activate(){
	if (textureID != TEXTURE_INVALID){
		glBindTexture(target, textureID);
		return true;
	} else return false;
}

bool Texture::deactivate(){
	return true;
}


#include "Texture2D.h"
#include "Texture3D.h"
#include "CubeMap.h"
#include "RenderTexture.h"
#include "ProceduralTexture.h"

/* ################################################################################################################### */

class TextureHandler {
private:
	Set <Texture *> textures;
	GLint nSimultaneousTextures;
    Set <GeneralFragmentShader *> fragmentShaders;
    Set <VertexShader *> vertexShaders;

	int currentTextureUnit;
	TextureID *currentTextures;
	TextureID *selectedTextures;
	
	TexEnv *currentTexEnvs;
	TexEnv *selectedTexEnvs;

    ShaderID selectedVertexShader,   currentVertexShader;
    ShaderID selectedFragmentShader, currentFragmentShader;

	GLenum selectedColorSrcFactor, selectedColorDestFactor;
	GLenum currentColorSrcFactor,  currentColorDestFactor;
	GLenum selectedAlphaSrcFactor, selectedAlphaDestFactor;
	GLenum currentAlphaSrcFactor,  currentAlphaDestFactor;
	GLenum currentBlendMode, selectedBlendMode;
	bool alphaBlendingEnabled;

	void reset();
	
public:
	TextureHandler();
	~TextureHandler();

	void setActiveTexture(int textureUnit);

	TextureID loadFromFile(const char *fileName);
    TextureID loadCubeMapFromFiles(const char *filePosX, const char *fileNegX, const char *filePosY, const char *fileNegY, const char *filePosZ, const char *fileNegZ);
    TextureID createNormalizingCubeMap(unsigned int size);
    TextureID createVectorField(unsigned int size);
    TextureID createAttenuationMap(unsigned int size);
    TextureID createSpecularMap(unsigned int width, unsigned int height, float minExponent, float maxExponent);
	TextureID createSpotLightMap(int size);

	TextureID loadFromMemory(unsigned char *pixels, unsigned int width, unsigned int height, unsigned int channels);
	TextureID createRenderTexture(unsigned int width, unsigned int height, bool wantCubeMap);
    TextureID createProceduralTexture(PROCEDURALTEXTUREFUNCTION proceduralFunc);
    bool update(TextureID texture, __int64 timeStamp);

	bool toNormalMap(TextureID texture);
	
//    bool setTextures(TextureID t0, TextureID t1, TextureID t2, TextureID t3, TextureID t4, TextureID t5, TextureID t6, TextureID t7, ...);

    bool setTextures(TextureID t0);
    bool setTextures(TextureID t0, TextureID t1);
    bool setTextures(TextureID t0, TextureID t1, TextureID t2);
    bool setTextures(TextureID t0, TextureID t1, TextureID t2, TextureID t3);

	bool setTexEnvs(TexEnv t0);
	bool setTexEnvs(TexEnv t0, TexEnv t1);
	bool setTexEnvs(TexEnv t0, TexEnv t1, TexEnv t2);
	bool setTexEnvs(TexEnv t0, TexEnv t1, TexEnv t2, TexEnv t3);

//	bool setTexEnvs(TexEnv t0, TexEnv t1, TexEnv t2, TexEnv t3, TexEnv t4, TexEnv t5, TexEnv t6, TexEnv t7, ...);

	bool setBlending(GLenum srcFactor, GLenum destFactor);
	bool setSeparateBlending(GLenum srcColorFactor, GLenum destColorFactor, GLenum srcAlphaFactor, GLenum destAlphaFactor);
    bool changeBlending(GLenum srcFactor, GLenum destFactor);
	bool changeBlending(GLenum colorSrcFactor, GLenum colorDestFactor, GLenum alphaSrcFactor, GLenum alphaDestFactor);
	bool setBlendMode(GLenum mode);
	bool apply();

	bool beginRenderToTexture(TextureID texture);
	bool endRenderToTexture(TextureID texture);
	bool setRenderFace(TextureID texture, GLint face);

    ShaderID beginVertexShader();
    void endVertexShader(ShaderID shader, GLuint firstInvariant);
    bool setVertexShader(ShaderID shader);
    GLuint getInvariants(ShaderID shader);

    ShaderID beginFragmentShader();
    void endFragmentShader(ShaderID shader);
    ShaderID beginRegisterCombiner();
    void endRegisterCombiner(ShaderID shader);

    bool setFragmentShader(ShaderID shader);

	bool upload(TextureID texture);
	bool uploadAll();
	bool unload(TextureID texture);
	bool unloadAll();
	void setFilter(TextureID texture, GLint filter);
	void setWrapMode(TextureID texture, GLint wrapModeS, GLint wrapModeT);
	void setWrapMode(TextureID texture, GLint wrapModeS, GLint wrapModeT, GLint wrapModeR);
    void setInternalFormat(TextureID texture, GLint internalFormat);
};

/* ----------------------------------------------------*/

void TextureHandler::reset(){
	for (int i = 0; i < nSimultaneousTextures; i++){
		selectedTextures [i] = TEXTURE_DISABLED;
		selectedTexEnvs  [i] = TEXENV_MODULATE;
	}

    selectedVertexShader   = SHADER_DISABLED;
    selectedFragmentShader = SHADER_DISABLED;

	selectedColorSrcFactor  = GL_ONE;
	selectedColorDestFactor = GL_ZERO;
	selectedAlphaSrcFactor  = GL_ONE;
	selectedAlphaDestFactor = GL_ZERO;
	selectedBlendMode  = GL_FUNC_ADD_EXT;
}

/* ----------------------------------------------------*/

TextureHandler::TextureHandler(){
	if (GL_ARB_multitexture){
		glGetIntegerv(GL_MAX_ACTIVE_TEXTURES_ARB, &nSimultaneousTextures);
	} else {
		nSimultaneousTextures = 1;
	}

    currentTextureUnit = 0;

	currentTextures  = new TextureID [nSimultaneousTextures];
	selectedTextures = new TextureID [nSimultaneousTextures];

	selectedTexEnvs  = new TexEnv    [nSimultaneousTextures];
	currentTexEnvs   = new TexEnv    [nSimultaneousTextures];

	reset();

    currentFragmentShader = SHADER_DISABLED;
    currentVertexShader   = SHADER_DISABLED;

	for (int i = 0; i < nSimultaneousTextures; i++){
		currentTextures [i] = TEXTURE_DISABLED;
		currentTexEnvs  [i] = TEXENV_MODULATE;
	}

	alphaBlendingEnabled = false;
	currentColorSrcFactor  = GL_ONE;
	currentColorDestFactor = GL_ZERO;
	currentAlphaSrcFactor  = GL_ONE;
	currentAlphaDestFactor = GL_ZERO;
	currentBlendMode  = GL_FUNC_ADD_EXT;
}

TextureHandler::~TextureHandler(){
	int i, len = textures.getSize();
	for (i = 0; i < len; i++){
		if (textures[i] != NULL){
			delete textures[i];
		}
	}

    len = fragmentShaders.getSize();
	for (i = 0; i < len; i++){
		if (fragmentShaders[i] != NULL){
			delete fragmentShaders[i];
		}
	}

    len = vertexShaders.getSize();
	for (i = 0; i < len; i++){
		if (vertexShaders[i] != NULL){
			delete vertexShaders[i];
		}
	}

	delete selectedTexEnvs;
	delete currentTexEnvs;

	delete selectedTextures;
	delete currentTextures;
}

void TextureHandler::setActiveTexture(int textureUnit){
	if (GL_ARB_multitexture && (textureUnit != currentTextureUnit)){
		glActiveTexture(GL_TEXTURE0_ARB + textureUnit);
		currentTextureUnit = textureUnit;
	}
}

TextureID TextureHandler::loadFromFile(const char *fileName){
	Texture2D *texture = new Texture2D();

	if (texture->loadFromFile(fileName)){
		int i, len = textures.getSize();

		for (i = 0; i < len; i++){
			if (textures[i] == NULL){
				textures[i] = texture;
				return i;
			}
		}
		return textures.add(texture);
	} else {
		delete texture;
		return TEXTURE_DISABLED;
	}	
}

TextureID TextureHandler::loadCubeMapFromFiles(const char *filePosX, const char *fileNegX, const char *filePosY, const char *fileNegY, const char *filePosZ, const char *fileNegZ){
	CubeMap *cubemap = new CubeMap();

	if (cubemap->loadFromFiles(filePosX, fileNegX, filePosY, fileNegY, filePosZ, fileNegZ)){
		int i, len = textures.getSize();

		for (i = 0; i < len; i++){
			if (textures[i] == NULL){
				textures[i] = cubemap;
				return i;
			}
		}
		return textures.add(cubemap);
	} else {
		delete cubemap;
		return TEXTURE_DISABLED;
	}	
}

TextureID TextureHandler::createNormalizingCubeMap(unsigned int size){
	CubeMap *cubemap = new CubeMap();

    cubemap->loadNormalizing(size);

    int i, len = textures.getSize();
    for (i = 0; i < len; i++){
        if (textures[i] == NULL){
			textures[i] = cubemap;
            return i;
        }
    }
    return textures.add(cubemap);
}

TextureID TextureHandler::createVectorField(unsigned int size){
    Texture3D *texture = new Texture3D();

    texture->createVectorField(size);
    int i, len = textures.getSize();
    for (i = 0; i < len; i++){
        if (textures[i] == NULL){
			textures[i] = texture;
            return i;
        }
    }
	return textures.add(texture);
}

TextureID TextureHandler::createAttenuationMap(unsigned int size){
    Texture3D *texture = new Texture3D();

    texture->createAttenuationMap(size);
    int i, len = textures.getSize();
    for (i = 0; i < len; i++){
        if (textures[i] == NULL){
            textures[i] = texture;
            return i;
        }
    }
	return textures.add(texture);
}

TextureID TextureHandler::createSpecularMap(unsigned int width, unsigned int height, float minExponent, float maxExponent){
	Texture2D *texture = new Texture2D();

    if (texture->createSpecularMap(width, height, minExponent, maxExponent)){
		int i, len = textures.getSize();

		for (i = 0; i < len; i++){
			if (textures[i] == NULL){
				textures[i] = texture;
				return i;
			}
		}
		return textures.add(texture);
	} else {
		delete texture;
		return TEXTURE_DISABLED;
	}	
}

TextureID TextureHandler::createSpotLightMap(int size){
	Texture2D *texture = new Texture2D();

    if (texture->createSpotLightMap(size)){
		int i, len = textures.getSize();

		for (i = 0; i < len; i++){
			if (textures[i] == NULL){
				textures[i] = texture;
				return i;
			}
		}
		return textures.add(texture);
	} else {
		delete texture;
		return TEXTURE_DISABLED;
	}
}

TextureID TextureHandler::loadFromMemory(unsigned char *pixels, unsigned int width, unsigned int height, unsigned int channels){
	Texture2D *texture = new Texture2D();

	texture->loadFromMemory(pixels, width, height, channels);
    int i, len = textures.getSize();
    for (i = 0; i < len; i++){
        if (textures[i] == NULL){
            textures[i] = texture;
            return i;
        }
    }
	return textures.add(texture);
}

TextureID TextureHandler::createRenderTexture(unsigned int width, unsigned int height, bool wantCubeMap){
    RenderTexture *texture = new RenderTexture(width, height, wantCubeMap);

    int i, len = textures.getSize();
    for (i = 0; i < len; i++){
        if (textures[i] == NULL){
            textures[i] = texture;
            return i;
        }
    }
	return textures.add(texture);
}

TextureID TextureHandler::createProceduralTexture(PROCEDURALTEXTUREFUNCTION proceduralFunc){
	ProceduralTexture *texture = new ProceduralTexture(proceduralFunc);

    int i, len = textures.getSize();
    for (i = 0; i < len; i++){
        if (textures[i] == NULL){
            textures[i] = texture;
            return i;
        }
    }
	return textures.add(texture);
}

bool TextureHandler::update(TextureID texture, __int64 timeStamp){
    return ((ProceduralTexture *) textures[texture])->update(timeStamp);
}


bool TextureHandler::toNormalMap(TextureID texture){
	return ((Texture2D *) textures[texture])->toNormalMap();
}


bool TextureHandler::setTextures(TextureID t0){
    selectedTextures[0] = t0;
    return true;
}

bool TextureHandler::setTextures(TextureID t0, TextureID t1){
    selectedTextures[0] = t0;
    selectedTextures[1] = t1;
    return true;
}

bool TextureHandler::setTextures(TextureID t0, TextureID t1, TextureID t2){
    selectedTextures[0] = t0;
    selectedTextures[1] = t1;
    selectedTextures[2] = t2;
    return true;
}

bool TextureHandler::setTextures(TextureID t0, TextureID t1, TextureID t2, TextureID t3){
    selectedTextures[0] = t0;
    selectedTextures[1] = t1;
    selectedTextures[2] = t2;
    selectedTextures[3] = t3;
    return true;
}
/*
bool TextureHandler::setTextures(TextureID t0 = TEXTURE_INVALID, TextureID t1 = TEXTURE_INVALID, TextureID t2 = TEXTURE_INVALID, TextureID t3 = TEXTURE_INVALID,
								 TextureID t4 = TEXTURE_INVALID, TextureID t5 = TEXTURE_INVALID, TextureID t6 = TEXTURE_INVALID, TextureID t7 = TEXTURE_INVALID, ...){
	va_list marker;
	int textureUnit;
	TextureID texture;
	
	texture = t0;
	textureUnit = 0;

	va_start(marker, t0);
	while (texture != TEXTURE_INVALID && textureUnit < nSimultaneousTextures){
		selectedTextures[textureUnit] = texture;

		texture = va_arg(marker, TextureID);
		textureUnit++;
	} 
	va_end(marker);

	return true;
}
*/

bool TextureHandler::setTexEnvs(TexEnv t0){
	selectedTexEnvs[0] = t0;
	return true;
}

bool TextureHandler::setTexEnvs(TexEnv t0, TexEnv t1){
	selectedTexEnvs[0] = t0;
	selectedTexEnvs[1] = t1;
	return true;
}

bool TextureHandler::setTexEnvs(TexEnv t0, TexEnv t1, TexEnv t2){
	selectedTexEnvs[0] = t0;
	selectedTexEnvs[1] = t1;
	selectedTexEnvs[2] = t2;
	return true;
}

bool TextureHandler::setTexEnvs(TexEnv t0, TexEnv t1, TexEnv t2, TexEnv t3){
	selectedTexEnvs[0] = t0;
	selectedTexEnvs[1] = t1;
	selectedTexEnvs[2] = t2;
	selectedTexEnvs[3] = t3;
	return true;
}

/*bool TextureHandler::setTexEnvs(TexEnv t0 = TEXENV_INVALID, TexEnv t1 = TEXENV_INVALID, TexEnv t2 = TEXENV_INVALID, TexEnv t3 = TEXENV_INVALID,
								TexEnv t4 = TEXENV_INVALID, TexEnv t5 = TEXENV_INVALID, TexEnv t6 = TEXENV_INVALID, TexEnv t7 = TEXENV_INVALID, ...){
	va_list marker;
	int textureUnit;
	TexEnv texEnv;
	
	texEnv = t0;
	textureUnit = 0;

	va_start(marker, t0);
	while (texEnv != TEXENV_INVALID && textureUnit < nSimultaneousTextures){
		selectedTexEnvs[textureUnit] = texEnv;

		texEnv = va_arg(marker, TexEnv);
		textureUnit++;
	} 
	va_end(marker);

	return true;
}*/

bool TextureHandler::setBlending(GLenum srcFactor, GLenum destFactor){
    selectedAlphaSrcFactor  = selectedColorSrcFactor  = srcFactor;
    selectedAlphaDestFactor = selectedColorDestFactor = destFactor;
	return true;
}

bool TextureHandler::setSeparateBlending(GLenum srcColorFactor, GLenum destColorFactor, GLenum srcAlphaFactor, GLenum destAlphaFactor){
	selectedColorSrcFactor  = srcColorFactor;
	selectedColorDestFactor = destColorFactor;
	selectedAlphaSrcFactor  = srcAlphaFactor;
	selectedAlphaDestFactor = destAlphaFactor;
    return true;
}

bool TextureHandler::changeBlending(GLenum srcFactor, GLenum destFactor){
	if (srcFactor != GL_ONE || destFactor != GL_ZERO){
		if (srcFactor != currentColorSrcFactor || destFactor != currentColorDestFactor ||
			srcFactor != currentAlphaSrcFactor || destFactor != currentAlphaDestFactor){
			glBlendFunc(srcFactor, destFactor);
			
			currentColorSrcFactor  = currentAlphaSrcFactor  = srcFactor;
			currentColorDestFactor = currentAlphaDestFactor = destFactor;
		}
		
		if (!alphaBlendingEnabled){
			glEnable(GL_BLEND);
			alphaBlendingEnabled = true;
		}
	} else {
		if (alphaBlendingEnabled){
			glDisable(GL_BLEND);
			alphaBlendingEnabled = false;
		}
	}
	return true;
}

bool TextureHandler::changeBlending(GLenum colorSrcFactor, GLenum colorDestFactor, GLenum alphaSrcFactor, GLenum alphaDestFactor){
	if (colorSrcFactor != alphaSrcFactor || colorDestFactor != alphaDestFactor){
		if (colorSrcFactor != currentColorSrcFactor || colorDestFactor != currentColorDestFactor ||
			alphaSrcFactor != currentAlphaSrcFactor || alphaDestFactor != currentAlphaDestFactor){
			glBlendFuncSeparateEXT(colorSrcFactor, colorDestFactor, alphaSrcFactor, alphaDestFactor);
				
			currentColorSrcFactor  = colorSrcFactor;
			currentAlphaSrcFactor  = alphaSrcFactor;
			currentColorDestFactor = colorDestFactor;
			currentAlphaDestFactor = alphaDestFactor;
		}
			
		if (!alphaBlendingEnabled){
			glEnable(GL_BLEND);
			alphaBlendingEnabled = true;
		}
		return true;
	} else return changeBlending(colorSrcFactor, colorDestFactor);
}

bool TextureHandler::setBlendMode(GLenum mode){
	selectedBlendMode = mode;
	return true;
}

bool TextureHandler::apply(){
    // Setup texture binding and texture enviroments
	for (int i = 0; i < nSimultaneousTextures; i++){
		if (selectedTextures[i] == TEXTURE_DISABLED){
			if (currentTextures[i] != TEXTURE_DISABLED){
                // Going from textured to non-textured
				setActiveTexture(i);
				textures[currentTextures[i]]->deactivate();
				glDisable(textures[currentTextures[i]]->getTarget());
				currentTextures[i] = TEXTURE_DISABLED;
			}
		} else {

			setActiveTexture(i);


			if (currentTextures[i] == TEXTURE_DISABLED){
				// Going from non-textured to textured
				glEnable(textures[selectedTextures[i]]->getTarget());
				textures[selectedTextures[i]]->activate();
				currentTextures[i] = selectedTextures[i];
			} else {
				if (currentTextures[i] != selectedTextures[i]){
					// Going from one texture to another texture
					textures[currentTextures[i]]->deactivate();

					int selectedTarget = textures[selectedTextures[i]]->getTarget();
					int currentTarget  = textures[currentTextures [i]]->getTarget();

					if (selectedTarget != currentTarget){
						glDisable(currentTarget);
						glEnable(selectedTarget);
					}
					textures[selectedTextures[i]]->activate();
					currentTextures[i] = selectedTextures[i];
                }
			}

			if (currentTexEnvs[i] != selectedTexEnvs[i]){
				switch(selectedTexEnvs[i]){
				case TEXENV_MODULATE:
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
					break;
				case TEXENV_MODULATE_NEXT:
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB,  GL_MODULATE);
					
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB,  GL_TEXTURE0_ARB + i);
					glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
					
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB,  GL_TEXTURE0_ARB + i + 1);
					glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
					glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_ARB, 1);
					break;
				case TEXENV_MODULATE_PREV_NEXT:
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB,  GL_MODULATE);
					
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB,  GL_PREVIOUS_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
					
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB,  GL_TEXTURE0_ARB + i + 1);
					glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
					glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_ARB, 1);
					break;
                case TEXENV_MODULATE_PRIMARY:
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB,  GL_MODULATE);
					
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB,  GL_PREVIOUS_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
					
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB,  GL_PRIMARY_COLOR_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
					glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_ARB, 1);
                    break;
                case TEXENV_REPLACE:
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
                    break;
				case TEXENV_ADD:
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_ADD);
					break;
				case TEXENV_ADD_PRIMARY:
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB,  GL_ADD);
					
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB,  GL_PREVIOUS_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
					
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB,  GL_PRIMARY_COLOR_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
					glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_ARB, 1);
					break;
                case TEXENV_INTERPOLATE:
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB,  GL_INTERPOLATE_ARB);
					
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB,  GL_PREVIOUS_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
					
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB,  GL_TEXTURE);
					glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);

                    glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_RGB_ARB,  GL_PRIMARY_COLOR_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_RGB_ARB, GL_SRC_COLOR);
					glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_ARB, 1);
                    break;
				case TEXENV_DOT3:
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB,  GL_DOT3_RGBA_ARB);
					
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB,  GL_PREVIOUS_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
					
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB,  GL_TEXTURE);
					glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
					glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_ARB, 1);
					break;
				case TEXENV_DOT3_NEXT:
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB,  GL_DOT3_RGBA_ARB);
					
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB,  GL_TEXTURE0_ARB + i);
					glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
					
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB,  GL_TEXTURE0_ARB + i + 1);
					glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
					glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_ARB, 1);
					break;
				}


				currentTexEnvs[i] = selectedTexEnvs[i];
			}
		}
	}

    // Setup vertex shaders
    if (selectedVertexShader == SHADER_DISABLED){
        if (currentVertexShader != SHADER_DISABLED){
            vertexShaders[currentVertexShader]->deactivate();
            glDisable(GL_VERTEX_SHADER_EXT);
            currentVertexShader = SHADER_DISABLED;
        }
    } else {
        if (currentVertexShader == SHADER_DISABLED){
            glEnable(GL_VERTEX_SHADER_EXT);
            vertexShaders[selectedVertexShader]->activate();
            currentVertexShader = selectedVertexShader;
        } else {
            if (selectedVertexShader != currentVertexShader){
                vertexShaders[currentVertexShader]->deactivate();
                vertexShaders[selectedVertexShader]->activate();
                currentVertexShader = selectedVertexShader;
            }
        }
    }

    // Setup fragment shaders
    if (selectedFragmentShader == SHADER_DISABLED){
        if (currentFragmentShader != SHADER_DISABLED){
            fragmentShaders[currentFragmentShader]->deactivate();
            glDisable(fragmentShaders[currentFragmentShader]->getFragmentShaderType());
            currentFragmentShader = SHADER_DISABLED;
        }
    } else {
        if (currentFragmentShader == SHADER_DISABLED){
            fragmentShaders[selectedFragmentShader]->activate();
            glEnable(fragmentShaders[selectedFragmentShader]->getFragmentShaderType());
            currentFragmentShader = selectedFragmentShader;
        } else {
            if (selectedFragmentShader != currentFragmentShader){
                fragmentShaders[currentFragmentShader]->deactivate();
                fragmentShaders[selectedFragmentShader]->activate();
                currentFragmentShader = selectedFragmentShader;
            }
        }
    }

    // Setup blending
/*	if (selectedColorSrcFactor != currentColorSrcFactor || selectedColorDestFactor != currentColorDestFactor ||
        selectedAlphaSrcFactor != currentAlphaSrcFactor || selectedAlphaDestFactor != currentAlphaDestFactor){
		if (selectedColorSrcFactor == BLEND_NONE){
			glDisable(GL_BLEND);
		} else {
            if (selectedColorSrcFactor == selectedAlphaSrcFactor && selectedColorDestFactor == selectedAlphaDestFactor){
    			glBlendFunc(selectedColorSrcFactor, selectedColorDestFactor);
            } else {
			    glBlendFuncSeparateEXT(selectedColorSrcFactor, selectedColorDestFactor, selectedAlphaSrcFactor, selectedAlphaDestFactor);
            }
			glEnable(GL_BLEND);
		}
		currentColorSrcFactor  = selectedColorSrcFactor;
		currentColorDestFactor = selectedColorDestFactor;
		currentAlphaSrcFactor  = selectedAlphaSrcFactor;
		currentAlphaDestFactor = selectedAlphaDestFactor;
	}*/

	changeBlending(selectedColorSrcFactor, selectedColorDestFactor, selectedAlphaSrcFactor, selectedAlphaDestFactor);

    // Setup blend mode
	if (selectedBlendMode != currentBlendMode){
		glBlendEquationEXT(selectedBlendMode);
		currentBlendMode = selectedBlendMode;
	}

	
	reset();

	return true;
}

bool TextureHandler::beginRenderToTexture(TextureID texture){
    // Hack to get it to work with different contexts

	for (int i = 0; i < nSimultaneousTextures; i++){
		currentTexEnvs[i] = TEXENV_INVALID;
	}

    reset();
    apply();

	currentColorSrcFactor = currentColorDestFactor = currentAlphaSrcFactor =  currentAlphaDestFactor = BLEND_INVALID;

	setActiveTexture(0);

	((RenderTexture *) textures[texture])->beginRenderToTexture();
	return true;
}

bool TextureHandler::endRenderToTexture(TextureID texture){
    // Hack to get it to work with different contexts
	for (int i = 0; i < nSimultaneousTextures; i++){
		currentTexEnvs[i] = TEXENV_INVALID;
	}

    reset();
    apply();

	currentColorSrcFactor = currentColorDestFactor = currentAlphaSrcFactor =  currentAlphaDestFactor = BLEND_INVALID;

	setActiveTexture(0);

	((RenderTexture *) textures[texture])->endRenderToTexture();
	return true;
}

bool TextureHandler::setRenderFace(TextureID texture, GLint face){
	((RenderTexture *) textures[texture])->setRenderFace(face);
	return true;
}

ShaderID TextureHandler::beginVertexShader(){
    VertexShader *shader = new VertexShader();

    shader->beginVertexShader();
    return vertexShaders.add(shader);
}

void TextureHandler::endVertexShader(ShaderID shader, GLuint firstInvariant){
    ((VertexShader *) vertexShaders[shader])->endVertexShader(firstInvariant);
}

bool TextureHandler::setVertexShader(ShaderID shader){
    selectedVertexShader = shader;
    return true;
}

GLuint TextureHandler::getInvariants(ShaderID shader){
    return ((VertexShader *) vertexShaders[shader])->getInvariants();
}

ShaderID TextureHandler::beginFragmentShader(){
    FragmentShader *shader = new FragmentShader();

    shader->beginFragmentShader();
    return fragmentShaders.add(shader);
}

void TextureHandler::endFragmentShader(ShaderID shader){
    ((FragmentShader *) fragmentShaders[shader])->endFragmentShader();
}

ShaderID TextureHandler::beginRegisterCombiner(){
    RegisterCombiner *shader = new RegisterCombiner();

    shader->beginFragmentShader();
    return fragmentShaders.add(shader);
}

void TextureHandler::endRegisterCombiner(ShaderID shader){
    ((RegisterCombiner *) fragmentShaders[shader])->endFragmentShader();
}

bool TextureHandler::setFragmentShader(ShaderID shader){
    selectedFragmentShader = shader;
    return true;
}


bool TextureHandler::upload(TextureID texture){
	return textures[texture]->upload();
}

bool TextureHandler::uploadAll(){
	int i, len = textures.getSize();
	for (i = 0; i < len; i++){
		textures[i]->upload();
	}
	return true;
}

bool TextureHandler::unload(TextureID texture){
	return textures[texture]->unload();
}

bool TextureHandler::unloadAll(){
	int i, len;

	for (i = 0; i < nSimultaneousTextures; i++){
		if (currentTextures[i] != TEXTURE_DISABLED){
			textures[currentTextures[i]]->deactivate();
			glDisable(textures[i]->getTarget());
			currentTextures[i] = TEXTURE_DISABLED;
		}
	}

	len = textures.getSize();
	for (i = 0; i < len; i++){
		textures[i]->unload();
	}
	return true;
}

void TextureHandler::setFilter(TextureID texture, GLint filter){
	textures[texture]->setFilter(filter);
}

void TextureHandler::setWrapMode(TextureID texture, GLint wrapModeS, GLint wrapModeT){
	textures[texture]->setWrapMode(wrapModeS, wrapModeT);
}

void TextureHandler::setWrapMode(TextureID texture, GLint wrapModeS, GLint wrapModeT, GLint wrapModeR){
	textures[texture]->setWrapMode(wrapModeS, wrapModeT, wrapModeR);
}

void TextureHandler::setInternalFormat(TextureID texture, GLint internalFormat){
    textures[texture]->setInternalFormat(internalFormat);
}



/* ################################################################################################################### */


#endif // _TEXTURE_H_